import { diff, fail, succeed, Tuple } from './fn.js'
import { ok } from '../vendor/kojima/src/index.js'
import { curry } from '../vendor/izuna/src/curry.js'

/** @import { ParseError, ParserState } from './state.js' */
/** @import { Result } from '../vendor/kojima/src/index.js' */

/**
 * @typedef {(value: any) => Result<ParserState, ParseError>} Parser
 */

/**
 * @param {...Parser} parsers
 */
export const seq = (...parsers) =>
  /** @param {ParserState} state */
  state => {
    let acc = ok(state)

    for (const parser of parsers) {
      if (acc.isOk()) {
        acc = acc.bind(parser)
      } else {
        break
      }
    }

    return acc
  }

export const map = curry(
  /** 
   * @param {(...args: any[]) => any} fn
   * @param {Parser} parser
   * @param {ParserState} state
   */
  (fn, parser, state) => {
    return parser(state).map(result => {
      try {
        /** @type {Result<ParserState, ParseError>} */
        const parsed = result.chain(otherState =>
          fn(diff(state[0], otherState[0]))
        )

        const backtrack = result.chain(otherState =>
          Tuple(state[0], otherState[1])
        )

        return succeed(parsed, backtrack)
      } catch (e) {
        return fail('failed to map', state, e)
      }
    })
  })

